# MIT License

# Copyright (c) 2018 AnonymousDapper

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# get_def options:

#  0  ->  crypto_box_SEEDBYTES
#  1  ->  crypto_pwhash_SALTBYTES
#  2  ->  crypto_pwhash_PASSWD_MIN
#  3  ->  crypto_pwhash_PASSWD_MAX
#  4  ->  crypto_pwhash_BYTES_MIN
#  5  ->  crypto_pwhash_BYTES_MAX
#  6  ->  crypto_pwhash_OPSLIMIT_MIN
#  7  ->  crypto_pwhash_OPSLIMIT_MAX
#  8  ->  crypto_pwhash_MEMLIMIT_MIN
#  9  ->  crypto_pwhash_MEMLIMIT_MAX
# 10  ->  crypto_secretbox_KEYBYTES
# 11  ->  crypto_secretbox_MACBYTES
# 12  ->  crypto_secretbox_NONCEBYTES
# 13  ->  crypto_secretbox_MESSAGEBYTES_MAX

"""
PyChickenTicket crypto module


:copyright: (c) 2018 AnonymousDapper
:license: MIT, see LICENSE for details.
"""

__title__ = "ct_crypto"
__author__ = "AnonymousDapper"
__license__ = "MIT"
__copyright__ = "Copyright 2018 AnonymousDapper"
__version__ = "0.0.62"


__all__ = ("encrypt", "random_bytes", "hexdigest", "genkey", "decrypt", "NONCE_BYTES", "SALT_BYTES", "KEY_BYTES", "LENGTH_BYTES")

from ._crypto_backend import get_def, genkey, random_bytes, hexdigest, derive_key, encrypt_secret, decrypt_secret

LENGTH_BYTES = 8
SALT_BYTES = get_def(1)
NONCE_BYTES = get_def(12)
KEY_BYTES = get_def(10)

PASS_MIN = get_def(2)
PASS_MAX = get_def(3)

MSG_MAX = get_def(13)


def encrypt(data: bytes, password: str):
    """Encrypt data with a password and return the encrypted data,
    packing the nonce and salt in as well.

    If data is encrypted with non-default length parameters,
    the decryption must be done with the same parameters.

    Parameters
    -----------
    data: bytes
        The binary data to encrypt.

    password: str
        The password to encrypt the data with.

    Returns
    --------
    bytes
        Bytes object containing the salt, nonce, and encrypted data

    Raises
    -------
    TypeError
        If argument datatypes are incorrect

    ValueError
        If argument length is incorrect

    RuntimeError
        If data encryption fails

    MemoryError
        If key derivation or encryption fails
    """

    if not isinstance(data, bytes):
        raise TypeError("expected bytes for data")

    elif len(data) > MSG_MAX:
        raise ValueError(f"data must be less than {MSG_MAX} bytes")

    if not isinstance(password, str):
        raise TypeError("expected str for password, not converting")

    elif not PASS_MIN < len(password) < PASS_MAX:
        raise ValueError(f"password must be between {PASS_MIN} and {PASS_MAX} chars")

    salt = random_bytes(SALT_BYTES)
    key = derive_key(password, salt, KEY_BYTES)

    nonce = random_bytes(NONCE_BYTES)
    encrypted = encrypt_secret(data, key, nonce)

    length = len(data).to_bytes(LENGTH_BYTES, byteorder="big")

    #print(f"Salt ({len(salt)} b): {hexdigest(salt)}")
    #print(f"Nonce ({len(nonce)} b): {hexdigest(nonce)}")
    #print(f"Key ({len(key)} b): {hexdigest(key)}")
    #print(f"Data ({len(encrypted)} b): {encrypted} - {hexdigest(encrypted)}")
    #print(f"Length: {length}")

    return salt + nonce + encrypted + length


def decrypt(data: bytes, password: str):
    """Decrypt data (assuming nonce and salt are packed in) with
    given password, and return decrypted data.

    If data is decrypted with non-default length parameters,
    the encryption must have been done with the same parameters.


    Parameters
    -----------
    data: bytes
        The binary data to encrypt.

    password: str
        The password to encrypt the data with.

    Returns
    --------
    bytes
        Bytes object containing the decrypted data.

    Raises
    -------
    TypeError
        If argument datatypes are incorrect

    ValueError
        If argument length is incorrect

    RuntimeError
        If data decryption authentication fails

    MemoryError
        If key derivation or decryption fails
    """

    if not isinstance(data, bytes):
        raise TypeError("expected bytes for data")

    if not isinstance(password, str):
        raise TypeError("expected str for password, not converting")

    elif not PASS_MIN < len(password) < PASS_MAX:
        raise ValueError(f"password must be between {PASS_MIN} and {PASS_MAX} chars")

    salt = data[:SALT_BYTES]
    nonce = data[SALT_BYTES:SALT_BYTES + NONCE_BYTES]
    encrypted = data[SALT_BYTES + NONCE_BYTES:-LENGTH_BYTES]
    length = data[-LENGTH_BYTES:]

    if (len(salt) != SALT_BYTES) or (len(nonce) != NONCE_BYTES) or (len(length) != LENGTH_BYTES):
        raise ValueError("data does not contained packed salt, nonce or length")

    if len(encrypted) > MSG_MAX:
        raise ValueError(f"data must be less than {MSG_MAX} bytes")

    key = derive_key(password, salt, KEY_BYTES)
    length = int.from_bytes(length, "big")

    #print(f"Salt ({len(salt)} b): {hexdigest(salt)}")
    #print(f"Nonce ({len(nonce)} b): {hexdigest(nonce)}")
    #print(f"Key ({len(key)} b): {hexdigest(key)}")
    #print(f"Data ({len(encrypted)} b): {encrypted} - {hexdigest(encrypted)}")
    #print(f"Length: {length}")

    decrypted = decrypt_secret(encrypted, length, key, nonce)

    return decrypted