/*
  MIT License

  Copyright (c) 2018 AnonymousDapper

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include <Python.h>
#include <sodium.h>

#include <string.h>


// Return value of requested definition
//
// get_def(def_: int)
//
// returns: def_value: int
static PyObject* _crypto_backend_get_def(PyObject *self, PyObject *args) {
  int def_request;
  unsigned long int def_value;

  if (!PyArg_ParseTuple(args, "i", &def_request)) {
    return NULL;
  }

  switch(def_request) {
    case 0:
      def_value = crypto_box_SEEDBYTES;
      break;

    case 1:
      def_value = crypto_pwhash_SALTBYTES;
      break;

    case 2:
      def_value = crypto_pwhash_PASSWD_MIN;
      break;

    case 3:
      def_value = crypto_pwhash_PASSWD_MAX;
      break;

    case 4:
      def_value = crypto_pwhash_BYTES_MIN;
      break;

    case 5:
      def_value = crypto_pwhash_BYTES_MAX;
      break;

    case 6:
      def_value = crypto_pwhash_OPSLIMIT_MIN;
      break;

    case 7:
      def_value = crypto_pwhash_OPSLIMIT_MAX;
      break;

    case 8:
      def_value = crypto_pwhash_MEMLIMIT_MIN;
      break;

    case 9:
      def_value = crypto_pwhash_MEMLIMIT_MAX;
      break;

    case 10:
      def_value = crypto_secretbox_KEYBYTES;
      break;

    case 11:
      def_value = crypto_secretbox_MACBYTES;
      break;

    case 12:
      def_value = crypto_secretbox_NONCEBYTES;
      break;

    case 13:
      def_value = crypto_secretbox_MESSAGEBYTES_MAX;
      break;

    default:
      PyErr_SetString(PyExc_KeyError, "the requested definition could not be found");
      return NULL;
  }

  PyObject *ret = Py_BuildValue("k", def_value);

  Py_INCREF(ret);

  return ret;
}

// Generate key N bytes long
//
// genkey()
//
// returns: key: bytes
static PyObject* _crypto_backend_genkey(PyObject *self, PyObject *args) {
  int key_len;

  if (!PyArg_ParseTuple(args, "i", &key_len)) {
    return NULL;
  }

  if (key_len < 32) {
    PyErr_SetString(PyExc_ValueError, "key length must be at least 32 bytes");
    return NULL;
  }

  unsigned char key[key_len];
  crypto_secretbox_keygen(key);

  PyObject *ret = Py_BuildValue("y#", key, sizeof key);

  return ret;
}

// Generate random buffer of N bytes
//
// random_bytes(size: int)
//
// returns: buf: bytes
static PyObject* _crypto_backend_random_bytes(PyObject *self, PyObject *args) {
  int buf_len;

  if (!PyArg_ParseTuple(args, "i", &buf_len)) {
    return NULL;
  }

  if (buf_len == 0)
    Py_RETURN_NONE;

  unsigned char buf[buf_len];
  randombytes_buf(buf, sizeof buf);

  PyObject *ret = Py_BuildValue("y#", buf, sizeof buf);

  Py_INCREF(ret);

  return ret;
}

// Return a hexdigest for byte
//
// hexdigest(data: bytes)
//
// returns: hexdigest: str
static PyObject* _crypto_backend_hexdigest(PyObject *self, PyObject *args) {
  Py_buffer input;

  if (!PyArg_ParseTuple(args, "y*", &input)) {
    return NULL;
  }

  if (input.len == 0)
    Py_RETURN_NONE;

  char *buf = (char *) input.buf;

  char *output = calloc(((input.len / sizeof(char)) * 2) + 1, sizeof(char));

  //char output[(strlen(buf) * 2) + 1];

  if (output == NULL) {
    PyErr_SetString(PyExc_MemoryError, "could not allocate memory for hexdigest buffer");
    return NULL;
  }

  for (int i = 0; i < input.len; i++) {
    char val = buf[i];
    char temp[3];

    sprintf(temp, "%02hhx", val);

    strcat(output, temp);

    //printf("Hex: %02hhx | Index: %i | Strlen: %lu\n", val, i, strlen(output));

  }

  PyObject *ret = Py_BuildValue("s", output);

  Py_INCREF(ret);

  return ret;
}

// Derive key from password and salt
//
// derive_key(password: str, salt: bytes, key_size: int)
//
// returns: key: bytes
static PyObject* _crypto_backend_derive_key(PyObject *self, PyObject *args) {
  int key_len;
  const char *password;
  Py_buffer salt_buf;
  unsigned char *salt;

  if (!PyArg_ParseTuple(args, "sy*i", &password, &salt_buf, &key_len))
    return NULL;

  if (key_len == 0 || salt_buf.len == 0)
    Py_RETURN_NONE;

  salt = (unsigned char *) salt_buf.buf;
  unsigned char key[key_len];

  if (crypto_pwhash(key, sizeof key, password, strlen(password), salt,
    crypto_pwhash_OPSLIMIT_INTERACTIVE, crypto_pwhash_MEMLIMIT_INTERACTIVE,
    crypto_pwhash_ALG_ARGON2ID13) != 0) {

    PyErr_SetString(PyExc_MemoryError, "Sodium key derivation failed, out of memory");
    return NULL;
  }

  PyObject *ret = Py_BuildValue("y#", key, sizeof key);

  Py_INCREF(ret);

  return ret;
}

// Encrypt data with key and nonce
//
// encrypt_secret(data: bytes, key: bytes, nonce: bytes)
//
// returns: encrypted_data: bytes
static PyObject* _crypto_backend_encrypt_secret(PyObject *self, PyObject *args) {
  Py_buffer nonce_buf;
  Py_buffer key_buf;
  Py_buffer input_buf;

  if (!PyArg_ParseTuple(args, "y*y*y*", &input_buf, &key_buf, &nonce_buf)) {
    return NULL;
  }

  if (nonce_buf.len == 0 || key_buf.len == 0 || input_buf.len == 0)
    Py_RETURN_NONE;

  unsigned char *nonce = (unsigned char *) nonce_buf.buf;
  unsigned char *key = (unsigned char *) key_buf.buf;
  unsigned char *input = (unsigned char *) input_buf.buf;

  int input_len = input_buf.len / sizeof(char);

  //printf("Length: %i\nMac bytes: %i\n", input_len, crypto_secretbox_MACBYTES);

  //unsigned char *output = calloc(crypto_secretbox_MACBYTES + input_len, sizeof(unsigned char));
  unsigned char output[crypto_secretbox_MACBYTES + input_len];

  if (output == NULL) {
    PyErr_SetString(PyExc_MemoryError, "could not allocate memory for encryption");
    return NULL;
  }

  if (crypto_secretbox_easy(output, input, input_len, nonce, key) != 0) {
    PyErr_SetString(PyExc_RuntimeError, "encryption failed");
    return NULL;
  }

  PyObject *ret = Py_BuildValue("y#", output, sizeof output);

  Py_INCREF(ret);

  return ret;
}

// Decrypt data with key and nonce
//
// decrypt_secret(data: bytes, msg_len: int, key: bytes, nonce:, bytes)
//
// returns decrypted_data: bytes
static PyObject* _crypto_backend_decrypt_secret(PyObject *self, PyObject *args) {
  Py_buffer nonce_buf;
  Py_buffer key_buf;
  Py_buffer input_buf;
  unsigned long long input_len;

  if (!PyArg_ParseTuple(args, "y*Ky*y*", &input_buf, &input_len, &key_buf, &nonce_buf)) {
    return NULL;
  }

  if (nonce_buf.len == 0 || key_buf.len == 0 || input_buf.len == 0)
    Py_RETURN_NONE;

  unsigned char *nonce = (unsigned char *) nonce_buf.buf;
  unsigned char *key = (unsigned char *) key_buf.buf;
  unsigned char *input = (unsigned char *) input_buf.buf;

  unsigned long long cipher_len = input_len + crypto_secretbox_MACBYTES;

  //unsigned char *output = calloc(input_len, sizeof(unsigned char));
  unsigned char output[input_len];

  if (output == NULL) {
    PyErr_SetString(PyExc_MemoryError, "could not allocate memory for decryption");
    return NULL;
  }

  if (crypto_secretbox_open_easy(output, input, cipher_len, nonce, key) != 0) {
    PyErr_SetString(PyExc_RuntimeError, "secret key decryption authentication failed");
    return NULL;
  }

  PyObject *ret = Py_BuildValue("y#", output, sizeof output);

  Py_INCREF(ret);

  return ret;
}

// Module setup

static PyMethodDef module_methods[] = {
  {"get_def", _crypto_backend_get_def, METH_VARARGS, "Get the value of a definition"},
  {"genkey", _crypto_backend_genkey, METH_VARARGS, "Generate a secret key of length N bytes"},
  {"random_bytes", _crypto_backend_random_bytes, METH_VARARGS, "Generate a random buffer of N bytes"},
  {"hexdigest", _crypto_backend_hexdigest, METH_VARARGS, "Return the hex digest"},
  {"derive_key", _crypto_backend_derive_key, METH_VARARGS, "Derive a secret key from a password and salt"},
  {"encrypt_secret", _crypto_backend_encrypt_secret, METH_VARARGS, "Encrypt data using a secret key and nonce"},
  {"decrypt_secret", _crypto_backend_decrypt_secret, METH_VARARGS, "Decrypt data using a secret key and nonce"},

  {NULL, NULL, 0, NULL}
};

static struct PyModuleDef module_def = {
  PyModuleDef_HEAD_INIT,
  "_crypto_backend",
  "Fast module for PyChickenTicket crypto",
  -1,
  module_methods
};

PyMODINIT_FUNC PyInit__crypto_backend(void) {
  Py_Initialize();

  if (sodium_init() < 0) {
    PyErr_SetString(PyExc_RuntimeError, "libsodium failed to initialize properly. This is not a ct_crypto error.");
    PyErr_Print();
    return NULL;
  }

  return PyModule_Create(&module_def);
}